from collections.abc import Sequence
from typing import TypeAlias

from mm_std import Err, Ok, Result, hr
from mm_std.random_ import random_str_choice
from pydantic import BaseModel

MAINNET_BASE_URL = "https://blockstream.info/api"
TESTNET_BASE_URL = "https://blockstream.info/testnet/api"

ERROR_INVALID_ADDRESS = "INVALID_ADDRESS"
ERROR_INVALID_NETWORK = "INVALID_NETWORK"

Proxy: TypeAlias = str | Sequence[str] | None


class Address(BaseModel):
    class ChainStats(BaseModel):
        funded_txo_count: int
        funded_txo_sum: int
        spent_txo_count: int
        spent_txo_sum: int
        tx_count: int

    class MempoolStats(BaseModel):
        funded_txo_count: int
        funded_txo_sum: int
        spent_txo_count: int
        spent_txo_sum: int
        tx_count: int

    chain_stats: ChainStats
    mempool_stats: MempoolStats


def get_address(
    address: str, testnet: bool = False, timeout: int = 10, proxies: Proxy = None, attempts: int = 1
) -> Result[Address]:
    result: Result[Address] = Err("not started yet")
    data = None
    base_url = TESTNET_BASE_URL if testnet else MAINNET_BASE_URL
    for _ in range(attempts):
        try:
            res = hr(f"{base_url}/address/{address}", timeout=timeout, proxy=random_str_choice(proxies))
            data = res.to_dict()
            if res.code == 400 and (
                "invalid bitcoin address" in res.body.lower() or "bech32 segwit decoding error" in res.body.lower()
            ):
                return Err(ERROR_INVALID_ADDRESS, data=data)
            elif res.code == 400 and "invalid network" in res.body.lower():
                return Err(ERROR_INVALID_NETWORK, data=data)
            return Ok(Address(**res.json))
        except Exception as err:
            result = Err(err, data=data)
    return result


def get_confirmed_balance(
    address: str, testnet: bool = False, timeout: int = 10, proxies: Proxy = None, attempts: int = 1
) -> Result[int]:
    return get_address(address, testnet=testnet, timeout=timeout, proxies=proxies, attempts=attempts).and_then(
        lambda a: Ok(a.chain_stats.funded_txo_sum - a.chain_stats.spent_txo_sum),
    )



from mm_btc.wallet import (
    BIP44_MAINNET_PATH,
    BIP44_TESTNET_PATH,
    BIP84_MAINNET_PATH,
    BIP84_TESTNET_PATH,
    derive_accounts,
    generate_mnemonic,
)


def test_generate_mnemonic():
    mnemonic = generate_mnemonic(strength=256)
    assert len(mnemonic.split()) == 24

    assert generate_mnemonic() != generate_mnemonic()


def test_derive_accounts():
    mnemonic = "arrange mutual earth tackle smart kangaroo rigid census exotic acoustic stock dream eager area laptop"
    passphrase = "my-secret"

    # mainnet bip44
    accs = derive_accounts(mnemonic, passphrase, BIP44_MAINNET_PATH, 2)
    assert accs[1].path == "m/44'/0'/0'/0/1"
    assert accs[1].address == "1L4Ghh4kuJuRCXrFrczwCV2TaggfCgf6B3"
    assert accs[1].private == "Kz4zoYPJ5astywoTbxHQr5SgFvMCekMqnWDFnXjY9CfvFtyNnGiE"

    # mainnet bip84
    accs = derive_accounts(mnemonic, passphrase, BIP84_MAINNET_PATH, 2)
    assert accs[1].path == "m/84'/0'/0'/0/1"
    assert accs[1].address == "bc1qrq2mwuqz3utyqz97xjq5npl0dyq5gmddpxdf8k"
    assert accs[1].private == "L447sW2JenqmEMMi7GoiUZDxTpY3Ni5ZfRrRxAbufiYwZV1AHytj"

    # testnet bip44
    accs = derive_accounts(mnemonic, passphrase, BIP44_TESTNET_PATH, 2)
    assert accs[1].path == "m/44'/1'/0'/0/1"
    assert accs[1].address == "n3DsvbopmKJaxHNgdZ6VVK4mB9fcuhXy5w"
    assert accs[1].private == "cVsAhYpSm2x8V2fQJdjmfBGm1aW7xKH1U8eZTpG9giz8usxNv4da"

    # testnet bip84
    accs = derive_accounts(mnemonic, passphrase, BIP84_TESTNET_PATH, 2)
    assert accs[1].path == "m/84'/1'/0'/0/1"
    assert accs[1].address == "tb1q4cu7pgs0w5s4ctjz4m6gx6m4rv8eqjwtnt0a4j"
    assert accs[1].private == "cPASMrozgCxXD6GQsJhge7hPY81j3huF8RJBY1PYwKmUVSYgeSRw"
